/*
Copyright (C)2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

static int DEBUG=0;
static char MODULE_NAME[]="bfcrypt";
static const int MAX_CIPHER_DATA_LENGTH=512;
static const char* PASSWORD_ENCRYPT_CONFIG_PROPERTY="com.buildforge.password.encryption.file";
static const char* DEFAULT_PASSWORD_ENCRYPT_CONFIG="bfpwcrypt.conf";
static const char* KEY_ALIAS_PROPERTY="bf_pwcrypt_key_alias";
static const char* KEY_PROPERTY="bf_pwcrypt_key";
static const char* KEY_DECRYPTED_PROPERTY="bf_pwcrypt_key_decrypted";
static const char* KEY_PASSWORD_PROPERTY="bf_pwcrypt_key_password";
static const char* KEY_PASSWORD_LOCATOR_PROPERTY="bf_pwcrypt_key_password_locator";
static hash_ctx *ctx = NULL;
static hash_table* ht_list[100];
static int num_keys=0;
static wchar_t swbuf[BFBUFSIZ];
static char scbuf[BFBUFSIZ];
static char trace_buf[512];

void BFTRACE_BFCRYPT (char *data) {
	FILE *file; 
	file = fopen("./bfcrypt.txt","a+"); /* apend file */
	fprintf(file,"%s",data);
	fclose(file);
}

#ifdef WIN32
#pragma data_seg(".GLOBALS")
int nProcessCount=0;
int nThreadCount=0;
#pragma data_seg()
BOOL WINAPI DllMain( HANDLE hinstDLL, DWORD ul_reason_for_call, LPVOID lpReserved) {
	switch (ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
		nProcessCount++;
		if (DEBUG) {
			sprintf(trace_buf, "Process count (incremented): %d\n", nProcessCount);
			BFTRACE_BFCRYPT(trace_buf);
		}
		break;
	case DLL_THREAD_ATTACH:
		nProcessCount--;
		if (DEBUG) {
			sprintf(trace_buf, "Process count (decremented): %d\n", nProcessCount);
			BFTRACE_BFCRYPT(trace_buf);
		}
		break;
	case DLL_THREAD_DETACH:
		nThreadCount++;
		if (DEBUG) {
			sprintf(trace_buf, "Thread count (incremented): %d\n", nThreadCount);
			BFTRACE_BFCRYPT(trace_buf);
		}
		break;
	case DLL_PROCESS_DETACH:
		nThreadCount--;
		if (DEBUG) {
			sprintf(trace_buf, "Thread count (decremented): %d\n", nThreadCount);
			BFTRACE_BFCRYPT(trace_buf);
		}
		break;
	}
	return TRUE;
}
#endif

/**
 *  Print an unsigned char* to a hexadecimal string
 **/
void printHexString(char* label, const unsigned char* key, const int key_length) {
	/* Just to be sure, don't print unless debug enabled. */
	if (DEBUG) {
		int i=0;
		unsigned char element;
		char trace[10];
		sprintf(trace_buf, "%s", label);
		for (i=0; i<key_length; i++) {
			element = key[i];
			sprintf(trace, "%02X ", element & 0xff);
			strncat(trace_buf, trace, 3);
		}
		strcat(trace_buf, "\n");
		BFTRACE_BFCRYPT(trace_buf);
	}
}

unsigned char* base64Decode (const char * data, int* ciphertext_length) {
#ifdef HAVE_SSL
	unsigned char *buffer;
    BIO *mem, *b64;
    int length=MAX_CIPHER_DATA_LENGTH;
	buffer = (unsigned char *)malloc(length);
    memset(buffer, 0, length);

	if (DEBUG) {
		sprintf(trace_buf, "base64Decode: Data before decode: %s\n", data);
		BFTRACE_BFCRYPT(trace_buf);
	}

    b64 = BIO_new(BIO_f_base64());
	BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
    mem = BIO_new_mem_buf((void*)data, strlen(data));
	mem = BIO_push(b64, mem);
	*ciphertext_length = BIO_read(mem, buffer, length);
	if (DEBUG) {
		sprintf(trace_buf, "base64Decode: Data read: %d\n", *ciphertext_length);
		BFTRACE_BFCRYPT(trace_buf);
	}
	BIO_flush(mem);
	BIO_free_all(mem);
	return buffer;
#else 
	return NULL;
#endif
}

char* base64Encode (const unsigned char* data, const int data_length) {
#ifdef HAVE_SSL
    BIO *mem, *b64;
    char* outbuff;
	int output_size;
	
	if (DEBUG) {
		sprintf(trace_buf, "base64Encode: Data length=%d\n", data_length);
		BFTRACE_BFCRYPT(trace_buf);
		printHexString("base64Encode: Encrypted data before encode: ", data, data_length);
	}

    outbuff = (char *)malloc(data_length);
    memset(outbuff, 0, data_length);
	b64 = BIO_new(BIO_f_base64());
	mem = BIO_new(BIO_s_mem());
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
	mem = BIO_push(b64, mem);
	BIO_write(mem, data, data_length);
    BIO_flush(mem);
	output_size = BIO_get_mem_data(mem, &outbuff);
	BIO_set_close(mem, BIO_NOCLOSE); 
	outbuff[output_size] = '\0';
	BIO_free_all(mem);

	if (DEBUG) {
		sprintf(trace_buf, "base64Encode: Encoded data=%s\n", outbuff);
		BFTRACE_BFCRYPT(trace_buf);
	}
	
	return outbuff;
#else
	return NULL;
#endif
}

static char *pt(unsigned char *md) {
	int i;
	static char buf[80];

	for (i=0; i<SHA_DIGEST_LENGTH; i++)
		sprintf(&(buf[i*2]),"%02x",md[i]);
	return(buf);
}

unsigned char* sha_digest (const char* decoded_key_password) {

	EVP_MD_CTX md;
    unsigned char *md_value;
	int md_len;
	unsigned char buf[1024];

	md_value = malloc (sizeof(unsigned char)*EVP_MAX_MD_SIZE);
	OpenSSL_add_all_digests();
    EVP_MD_CTX_init(&md);
    EVP_DigestInit_ex(&md, EVP_sha1(), NULL);
	strcpy((char *)buf, (const char *)decoded_key_password);
	EVP_DigestUpdate(&md, buf, strlen(decoded_key_password));
	if(EVP_DigestFinal_ex(&md, md_value, (unsigned int*)(&md_len)) != 1) {
		if (DEBUG) {
			sprintf(trace_buf, "sha_digest: Digest computation problem\n");
			BFTRACE_BFCRYPT(trace_buf);
		}
		/* TODO: Need to link send_msg into the DLL w/o bringing in the kitchen sink.  
		The message is completed though.
		send_msg("PWCrypt", "PWCryptDigestError", NULL);
		*/
		return NULL;
	}
    EVP_MD_CTX_cleanup(&md);
	return md_value;
}

/* Used to pad the input data of openssl_aes_cbc_encrypt. */
static int pad(unsigned char *data, int data_len, int block_size, unsigned char **padded_data) {
	int modulus=0, i=0;
	char pad_char;
	char* pad_holder;

	/* determine the data_len modulus of the block size. */
	modulus = block_size - (data_len%block_size);

	if (DEBUG) {
		sprintf (trace_buf, "pad: Data length=%d, Block size=%d, Modulus=%d\n", data_len, block_size, modulus);
		BFTRACE_BFCRYPT(trace_buf);
	}

	if (modulus == block_size) {
		*padded_data = data;
		return data_len;
	}

	*padded_data = malloc(data_len+modulus+1);
	memset(*padded_data, 0, data_len+modulus);
	memcpy(*padded_data, data, data_len);

	pad_char = (char)modulus&0xff;
	pad_holder = (char *)malloc(block_size+1);

	for (i=0; i<block_size; i++) {
		pad_holder[i]=(char)pad_char;
	}
	pad_holder[block_size]='\0';

	if (DEBUG) {
		sprintf (trace_buf, "pad: Pad character = %d, Pad holder = %s\n", pad_char, pad_holder);
		BFTRACE_BFCRYPT(trace_buf);
	}

	/* add 01 if modulus 1, add 02 if modulus 2, etc. */
	for (i=0; i<modulus; i++) {
		memcpy(*padded_data+data_len, pad_holder, modulus);
	}
	
	padded_data[data_len+modulus]='\0';
	return (data_len+modulus);
}

/**
 * The internal encrypt method that handles AES encryption.
 */
int encrypt_AES128CBC (const unsigned char* decoded_key, const unsigned char* iv, const char* plaintext, int plaintext_length, unsigned char** ciphertext, int* ciphertext_length) {
#ifdef HAVE_SSL    
	EVP_CIPHER_CTX ctx;
	const EVP_CIPHER *aes;
	int SALT_LENGTH = 10, i=0, length=0, out_length=0, num_chunks=0, currentReadLength=0, currentDataLength=0;
	unsigned char *random_plus_plaintext=NULL, random_data[11], *encrypt_buffer=NULL;
	
	/* create the AES 128 CBC cipher. */
	aes=EVP_aes_128_cbc();
	if (DEBUG) {
		sprintf(trace_buf, "encrypt_AES128CBC: Encrypt block size: %d\n", aes->block_size);
		BFTRACE_BFCRYPT(trace_buf);
	}
	
	encrypt_buffer = malloc(512);
	memset(encrypt_buffer, 0, 512);

	/* generate SALT */
	RAND_pseudo_bytes(random_data, SALT_LENGTH+1);
	for (i=0; i<SALT_LENGTH; i++) {
		if (random_data[i] == 0) random_data[i]++;
	}

	// random_data = sha_digest(plaintext);
	random_data[SALT_LENGTH]='\0';
	
	/* Concat random_data with plaintext data then PKCS#5 pad the data to an even block boundary */
	length = SALT_LENGTH + plaintext_length;
	random_plus_plaintext = malloc(length+1);
	memset(random_plus_plaintext, 0, length+1);
	memcpy(random_plus_plaintext, random_data, SALT_LENGTH);
	memcpy(random_plus_plaintext+SALT_LENGTH, plaintext, plaintext_length);
	length = pad(random_plus_plaintext, length, aes->block_size, &random_plus_plaintext);
	random_plus_plaintext[length]='\0';

	if (DEBUG) {
		printHexString("encrypt_AES128CBC: Random/Data (hex): ", random_plus_plaintext, length);
		sprintf(trace_buf, "encrypt_AES128CBC: Length: %d\n", length);
		BFTRACE_BFCRYPT(trace_buf);
	}

	/* Start initialization */
	EVP_CIPHER_CTX_init(&ctx);
	EVP_CIPHER_CTX_set_padding(&ctx, 0);

	if(!EVP_EncryptInit_ex(&ctx,aes,NULL,decoded_key,iv)) {
		if (DEBUG) {
			sprintf(trace_buf, "encrypt_AES128CBC: Encrypt initialization failed.\n");
			BFTRACE_BFCRYPT(trace_buf);
		}
		/* TODO: Need to link send_msg into the DLL w/o bringing in the kitchen sink.  
		The message is completed though.
		send_msg("PWCrypt", "PWCryptEncryptInitFailed", NULL);
		*/
		return -1;
	}
	
	num_chunks=length/aes->block_size;
	if (DEBUG) {
		sprintf(trace_buf, "encrypt_AES128CBC: Num chunks: %d\n", num_chunks);
		BFTRACE_BFCRYPT(trace_buf);
	}

	if (length>=aes->block_size)
		for (i=0; i<num_chunks; i++) {
			if(!EVP_EncryptUpdate(&ctx,&encrypt_buffer[currentDataLength],&out_length,&random_plus_plaintext[currentReadLength],aes->block_size)) {
				if (DEBUG) {
					sprintf(trace_buf, "encrypt_AES128CBC: Encrypt length: %d\n", out_length);
					BFTRACE_BFCRYPT(trace_buf);
				}
				ERR_print_errors_fp(stderr);
				return -1;
			}
			currentDataLength += out_length;
			currentReadLength += aes->block_size;
			if (DEBUG) {
				sprintf(trace_buf, "encrypt_AES128CBC: Current data length: %d\n", currentDataLength);
				BFTRACE_BFCRYPT(trace_buf);
				sprintf(trace_buf, "encrypt_AES128CBC: Current read length: %d\n", currentReadLength);
				BFTRACE_BFCRYPT(trace_buf);
			}
		}

	if (length%aes->block_size) {
		if(!EVP_EncryptUpdate(&ctx,&encrypt_buffer[currentDataLength],&out_length,&random_plus_plaintext[currentReadLength],aes->block_size)) {
			if (DEBUG) {
				sprintf(trace_buf, "encrypt_AES128CBC: Encrypt mod length: %d\n", out_length);
				BFTRACE_BFCRYPT(trace_buf);
			}
			ERR_print_errors_fp(stderr);
			return -1;
		}
		currentDataLength += out_length;
		currentReadLength += aes->block_size;
		if (DEBUG) {
			sprintf(trace_buf, "encrypt_AES128CBC: Current data length after mod update: %d\n", currentDataLength);
			BFTRACE_BFCRYPT(trace_buf);
			sprintf(trace_buf, "encrypt_AES128CBC: Current read length after mod update: %d\n", currentReadLength);
			BFTRACE_BFCRYPT(trace_buf);
		}
	}

	if(!EVP_EncryptFinal_ex(&ctx,&encrypt_buffer[currentDataLength],&out_length)) {
		if (DEBUG) { 
			sprintf(trace_buf, "encrypt_AES128CBC: Final encrypt length: %d\n", out_length);
			BFTRACE_BFCRYPT(trace_buf);
			sprintf(trace_buf, "encrypt_AES128CBC: Ciphertext: %s\n", encrypt_buffer);
			BFTRACE_BFCRYPT(trace_buf);
		}
	    ERR_print_errors_fp(stderr);
	    return -1;
	}
	
	currentDataLength += out_length;
	currentReadLength += aes->block_size;

	if (DEBUG) {
		sprintf(trace_buf, "encrypt_AES128CBC: Final encrypt length: %d\n", currentDataLength);
		BFTRACE_BFCRYPT(trace_buf);
		sprintf(trace_buf, "encrypt_AES128CBC: Final read length: %d\n", currentReadLength);
		BFTRACE_BFCRYPT(trace_buf);
		sprintf(trace_buf, "encrypt_AES128CBC: Ciphertext: %s\n", encrypt_buffer);
		BFTRACE_BFCRYPT(trace_buf);
	}

	EVP_CIPHER_CTX_cleanup(&ctx);

	/* put the data in the outbound pointer */
	*ciphertext = malloc((sizeof(unsigned char)*currentDataLength)+1);
	*ciphertext_length = currentDataLength;
	memset(*ciphertext, 0, currentDataLength+1);
	memcpy(*ciphertext, encrypt_buffer, currentDataLength);
	

	if (DEBUG) {
		sprintf(trace_buf, "encrypt_AES128CBC: Encrypted data length: %d\n", *ciphertext_length);
		BFTRACE_BFCRYPT(trace_buf);
		printHexString("encrypt_AES128CBC: Encrypted data: ", *ciphertext, *ciphertext_length);
	}

	if (encrypt_buffer) free(encrypt_buffer);
	if (random_plus_plaintext) free(random_plus_plaintext);
	return 0;	
#else /* !HAVE_SSL */
	return -1;
#endif
}

int decrypt_AES128CBC (const unsigned char* decoded_key, const unsigned char* iv, unsigned char** plaintext, const unsigned char* ciphertext, const int ciphertext_length, const int remove_padding) {
#ifdef HAVE_SSL    
	EVP_CIPHER_CTX ctx;
	const EVP_CIPHER *aes;
	unsigned char buffer[512];
	int SALT_LENGTH = 10, i=0, out_length=0, out_length_final=0, num_chunks=0, currentReadLength=0, currentDataLength=0;

	memset(buffer, 0, 512);

	if (DEBUG) {
		sprintf(trace_buf, "decrypt_AES128CBC: Ciphertext length: %d\n", ciphertext_length);
		BFTRACE_BFCRYPT(trace_buf);
		printHexString("decrypt_AES128CBC: Ciphertext data: ", ciphertext, ciphertext_length);
	}

	/* create the AES 128 CBC cipher. */
	aes=EVP_aes_128_cbc();
	
	EVP_CIPHER_CTX_init(&ctx);

	if(!EVP_DecryptInit_ex(&ctx,aes,NULL,decoded_key,iv)) {
	    ERR_print_errors_fp(stderr);
		return -1;
	}
	
	num_chunks=ciphertext_length/aes->block_size;
	
	if (DEBUG) {
		sprintf(trace_buf, "decrypt_AES128CBC: Decrypt block size: %d\n", aes->block_size);
		BFTRACE_BFCRYPT(trace_buf);
		sprintf(trace_buf, "decrypt_AES128CBC: Ciphertext length: %d\n", ciphertext_length);
		BFTRACE_BFCRYPT(trace_buf);
		sprintf(trace_buf, "decrypt_AES128CBC: Num chunks: %d\n", num_chunks);
		BFTRACE_BFCRYPT(trace_buf);
	}
		
	if (num_chunks > 2) {
		for (i=0; i<num_chunks; i++) {
			if(!EVP_DecryptUpdate(&ctx,&buffer[currentDataLength],&out_length,&ciphertext[currentReadLength],aes->block_size)) {
				if (DEBUG) {
					sprintf(trace_buf, "decrypt_AES128CBC: Decrypt length: %d\n", out_length);
					BFTRACE_BFCRYPT(trace_buf);
				}
				return -2;
			}
			currentDataLength += out_length;
			currentReadLength += aes->block_size;
			if (DEBUG) {
				sprintf(trace_buf, "decrypt_AES128CBC: Current data length: %d\n", currentDataLength);
				BFTRACE_BFCRYPT(trace_buf);
				sprintf(trace_buf, "decrypt_AES128CBC: Current read length: %d\n", currentReadLength);
				BFTRACE_BFCRYPT(trace_buf);
				sprintf(trace_buf, "decrypt_AES128CBC: Decrypt length: %d\n", out_length);
				BFTRACE_BFCRYPT(trace_buf);
			}
		}

		if (ciphertext_length%aes->block_size) {
			if(!EVP_DecryptUpdate(&ctx,&buffer[currentReadLength],&out_length,&ciphertext[currentDataLength], ciphertext_length%aes->block_size)) {
				return -3;
			}
	        
			currentDataLength += out_length;
			currentReadLength += ciphertext_length%aes->block_size;

			if (DEBUG) {
				sprintf(trace_buf, "decrypt_AES128CBC: Current data length: %d\n", currentDataLength);
				BFTRACE_BFCRYPT(trace_buf);
				sprintf(trace_buf, "decrypt_AES128CBC: Current read length: %d\n", currentReadLength);
				BFTRACE_BFCRYPT(trace_buf);
			}
		}

		if(out_length > 0) {
			if (!EVP_DecryptFinal_ex(&ctx,buffer+out_length,&out_length_final)) {
				if (DEBUG) {
					sprintf(trace_buf, "decrypt_AES128CBC: Final decrypt length: %d\n", out_length);
					BFTRACE_BFCRYPT(trace_buf);
				}
			}
			currentDataLength += out_length;
			currentReadLength += ciphertext_length%aes->block_size;
		}

		if (DEBUG) {
			sprintf(trace_buf, "decrypt_AES128CBC: Final decrypt length: %d\n", out_length_final);
			BFTRACE_BFCRYPT(trace_buf);
		}
	} else {
		if (!EVP_DecryptUpdate(&ctx,buffer,&out_length,ciphertext,ciphertext_length)) {
			if (DEBUG) {
				sprintf(trace_buf, "decrypt_AES128CBC: Decrypt length: %d\n", out_length);
				BFTRACE_BFCRYPT(trace_buf);
			}
			return -4;
		}
		currentDataLength += out_length;
		currentReadLength += ciphertext_length%aes->block_size;

		if (out_length > 0) {
			if (!EVP_DecryptFinal(&ctx,buffer+out_length,&out_length_final)) {
				if (DEBUG) {
 					sprintf(trace_buf, "decrypt_AES128CBC: Final decrypt length: %d\n", out_length_final);
					BFTRACE_BFCRYPT(trace_buf);
					sprintf(trace_buf, "decrypt_AES128CBC: Final out length: %d\n", out_length);
					BFTRACE_BFCRYPT(trace_buf);
				}
			}
			currentDataLength += out_length_final;
			currentReadLength += ciphertext_length%aes->block_size;
		}
	}

	currentDataLength = ciphertext_length; 
	if (remove_padding == 1) {
		out_length_final = currentDataLength;
		for (i=out_length_final-1; i>9; i--) {
			if ((unsigned char)buffer[i] < 0x20) {
				buffer[i] = 0;
				currentDataLength = i;
			}
			if (buffer[i] == 0) {
				currentDataLength = i;
			}
		}
	}

	*plaintext = malloc(sizeof(unsigned char)*(ciphertext_length));
	memset(*plaintext, 0, ciphertext_length);

	/* remove the first SALT_LENGTH from the beginning of the data */
	memcpy(*plaintext, buffer+SALT_LENGTH, ciphertext_length-SALT_LENGTH);

	if (DEBUG && remove_padding) {
		/* if the final decrypt fails, go with the intermediate results. */
		printHexString("decrypt_AES128CBC: Data: ", *plaintext, strlen((const char*)*plaintext));
		sprintf(trace_buf, "decrypt_AES128CBC: out_length_final: %d\n", out_length_final);
		BFTRACE_BFCRYPT(trace_buf);
		sprintf(trace_buf, "decrypt_AES128CBC: currentDataLength: %d\n", currentDataLength);
		BFTRACE_BFCRYPT(trace_buf);
	}

	EVP_CIPHER_CTX_cleanup(&ctx);
	return 0;
#else /* !HAVE_SSL */
	return -5;
#endif
}

/**
 * This is used to decrypt the master key.  Using the SHA of the decoded_key_password as the decryption key.
 */
unsigned char* master_key_decrypt(const unsigned char* decoded_key, const int key_length, const unsigned char* decoded_key_password) {
	unsigned char* digest;
	unsigned char key[16], iv16[16];
	unsigned char *decrypted_key=NULL;
	int rc=0;

	digest = sha_digest((const char *)decoded_key_password);
	memcpy(key, digest, 16);
	memcpy(iv16, digest, 16);
	free (digest); // free the memory held by digest.

	rc = decrypt_AES128CBC ((const unsigned char*)key, (const unsigned char*)iv16, &decrypted_key, decoded_key, key_length, 0);

	if (rc != 0) {
		if (DEBUG) {
			sprintf(trace_buf, "master_key_decrypt: Problem decoding master key.  Error: %d\n", rc);
			BFTRACE_BFCRYPT(trace_buf);
		}
	}
	
	return decrypted_key;
}

char* getName(void) {
	return (char*)&MODULE_NAME;
}

int encrypt_password (const char* plaintext, unsigned char** ciphertext, int* ciphertext_length, char** info) {
	int i=0, rc=0, key_length=0, pwd_length=0;
	hash_table *ht = NULL;
	char *key_password_value=NULL, *key_password_loc_value=NULL;
	unsigned char *decoded_key_password=NULL, *decoded_key=NULL, *key_value=NULL;
	char buf[256];

	if (DEBUG) {
		sprintf(trace_buf, "encrypt: num_keys: %d\n", num_keys);
		BFTRACE_BFCRYPT(trace_buf);
	}
	/* Get the latest key map to encrypt with. */
	for (i=num_keys-1; i>=0; i--) {
		ht = ht_list[i];
		
		/* This is the data that gets returned for the password header. */
		*info = hash_get_str(ht, KEY_ALIAS_PROPERTY);

		if (DEBUG) {
			sprintf(trace_buf, "encrypt: info=%s\n", *info);
			BFTRACE_BFCRYPT(trace_buf);
			sprintf(trace_buf, "encrypt: info length=%d\n", strlen(*info));
			BFTRACE_BFCRYPT(trace_buf);
		}
		
		/* Get the decrypted key from the key map, if exists. */
		key_value = hash_get_str(ht, KEY_DECRYPTED_PROPERTY);

		if (key_value == NULL) {
			/* We need the encrypted key now. */
			key_value = hash_get_str(ht, KEY_PROPERTY);
			if (key_value != NULL) {
				/* Decode it first. */
				decoded_key = base64Decode((const char*)key_value, &key_length);
								
				// TODO: Handle password locator in bfpwcrypt.conf (save this nice piece of work for a little later).
				key_password_loc_value = hash_get_str(ht, KEY_PASSWORD_LOCATOR_PROPERTY);

				if (DEBUG && key_password_loc_value != NULL) {
					sprintf(trace_buf, "encrypt: Key password locator: %s\n", key_password_loc_value);
					BFTRACE_BFCRYPT(trace_buf);
				}

				key_password_value = hash_get_str(ht, KEY_PASSWORD_PROPERTY);
				decoded_key_password = base64Decode(key_password_value, &pwd_length);
				key_value = master_key_decrypt(decoded_key, key_length, decoded_key_password);
				if (DEBUG) printHexString("encrypt: Key debug data (first 8 bytes): ", key_value, 8); 
				hash_put_str(ht, (char*)KEY_DECRYPTED_PROPERTY, key_value);
				break; 
			} 
		} else {
			if (DEBUG) printHexString("encrypt: Key debug data (first 8 bytes): ", key_value, 8); 
		}
	}
	
	if (DEBUG) {
		sprintf(trace_buf, "encrypt: info before encrypt: %s\n", *info);
		BFTRACE_BFCRYPT(trace_buf);
	}

	memset(buf, 0, 256);
	memcpy(buf, *info, strlen(*info));
	
	rc = encrypt_AES128CBC(key_value, key_value, plaintext, strlen(plaintext), ciphertext, ciphertext_length);

	if (*info == NULL) *info = (char*)&buf;

	if (DEBUG) {
		sprintf(trace_buf, "encrypt: info after encrypt: %s\n", *info);
		BFTRACE_BFCRYPT(trace_buf);
		sprintf(trace_buf, "encrypt: Return Code: %d\n", rc);
		BFTRACE_BFCRYPT(trace_buf);
	}
	
	if (rc != 0) {
		if (DEBUG) {
			sprintf(trace_buf, "encrypt: Failure encrypting password. Error: %d\n", rc);
			BFTRACE_BFCRYPT(trace_buf);
		}
		/* TODO: Need to link send_msg into the DLL w/o bringing in the kitchen sink.  
		The message is completed though.
		send_msg("PWCrypt", "PWCryptEncryptFailed", "i", rc);
		*/
	}

	return rc;
}

int decrypt_password (char** plaintext, const unsigned char* ciphertext, const int ciphertext_length, const char* info) {
	int i=0, key_length=0, pwd_length=0, found=0;
	hash_table *ht = NULL;
	char *info_value=NULL, *key_value=NULL, *key_password_value=NULL, *key_password_loc_value=NULL;
	unsigned char *decoded_key_password=NULL, *decoded_key=NULL, *decrypted_key=NULL;

	/* info to be used to get access to the proper key. */
	if (DEBUG) {
		sprintf(trace_buf, "decrypt: Looking for key matching info: %s, length=%d\n", info, strlen(info));
		BFTRACE_BFCRYPT(trace_buf);
	}

	for (i=0; i<num_keys; i++) {
		ht = ht_list[i]; 
		info_value = hash_get_str(ht, KEY_ALIAS_PROPERTY);
		if (info_value == NULL) {
			if (DEBUG) {
				sprintf(trace_buf, "decrypt: Info value is null for ht[%d]\n", i);
				BFTRACE_BFCRYPT(trace_buf);
			}
			continue;
		}
		
		if (DEBUG) {
			sprintf(trace_buf, "decrypt: Comparing against: %s, length=%d\n", info_value, strlen(info_value));
			BFTRACE_BFCRYPT(trace_buf);
		}

		/* if the key alias matches the password info, then use that key to decrypt. */
		if (strncmp(info_value, info, strlen(info_value)) == 0) {
			if (DEBUG) {
				sprintf(trace_buf, "decrypt: Found match: %s, length=%d\n", info_value, strlen(info_value));
				BFTRACE_BFCRYPT(trace_buf);
			}
			decrypted_key = hash_get_str(ht, KEY_DECRYPTED_PROPERTY);
			
			/* decrypt the key if not already done. */
			if (decrypted_key == NULL) {
				/* Get the encrypted key */
				key_value = hash_get_str(ht, KEY_PROPERTY);
				if (key_value != NULL) {
					/* Decode it first. */
					decoded_key = base64Decode(key_value, &key_length);
					// TODO: Handle password locator in bfpwcrypt.conf (save this nice piece of work for a little later).
					key_password_loc_value = hash_get_str(ht, KEY_PASSWORD_LOCATOR_PROPERTY);
					if (DEBUG && key_password_loc_value != NULL) {
						sprintf(trace_buf, "decrypt: Key password locator: %s\n", key_password_loc_value);
						BFTRACE_BFCRYPT(trace_buf);
					}
	
					key_password_value = hash_get_str(ht, KEY_PASSWORD_PROPERTY);
					decoded_key_password = base64Decode(key_password_value, &pwd_length);
					decrypted_key = master_key_decrypt(decoded_key, key_length, decoded_key_password);
					hash_put_str(ht, (char*)KEY_DECRYPTED_PROPERTY, decrypted_key);
				}
			} else {
				if (DEBUG) {
					sprintf(trace_buf, "decrypt: Key already decrypted.\n");
					BFTRACE_BFCRYPT(trace_buf);
					printHexString("decrypt: Key debug data (first 8 bytes): ", decrypted_key, 8); 
				}
			}
			found=1;
			break;
		} 
	}

	if (!found) {
		if (DEBUG) {
			sprintf(trace_buf, "decrypt: Warning!  No matching key found.\n");
			BFTRACE_BFCRYPT(trace_buf);
		}
		/* TODO: Need to link send_msg into the DLL w/o bringing in the kitchen sink.  
		The message is completed though.
		send_msg("PWCrypt", "PWCryptNoMatchingKeyForAlias", "s", info);
		*/
		return -1;
	}

	return decrypt_AES128CBC((const unsigned char*)decrypted_key, (const unsigned char*)decrypted_key, (unsigned char**)plaintext, ciphertext, ciphertext_length, 1);
}

passwd_ops my_ops = {
		&getName,
        &encrypt_password,
        &decrypt_password
};

/**
 * This should be in bfcrypt.c, writing it here to get some function moving along.
 **/
int load_keys_from_file (char * params) {
	char	*password_encrypt_config = NULL;
	FILE    *f;
	char    *buf, *key, *value, *separator_ptr;
	int 	j=0, i=0, strings=1, key_size=0, value_size=0, total_size=0;
	hash_linked_list *list;
	hash_table *ht = NULL;

	if (DEBUG) {
		sprintf(trace_buf, "load_keys_from_file: Entry -> load_key_file.\n");
		BFTRACE_BFCRYPT(trace_buf);
	}
	if (ctx == NULL) {
		ctx = hash_ctx_new();
	}

	if (params == NULL) {
		/* open the file if it exists from the config */
		password_encrypt_config = getenv(PASSWORD_ENCRYPT_CONFIG_PROPERTY);

		/* default if one is not specified */
		if (password_encrypt_config == NULL) {
			password_encrypt_config = (char*)DEFAULT_PASSWORD_ENCRYPT_CONFIG;
		}

		if (password_encrypt_config == NULL) {
			if (DEBUG) {
				sprintf(trace_buf, "load_keys_from_file: Could not find password encrypt configuration: %s\n", DEFAULT_PASSWORD_ENCRYPT_CONFIG);
				BFTRACE_BFCRYPT(trace_buf);
			}
			return -1;
		}
	} else {
		password_encrypt_config = params;
	}

	if (DEBUG) {
		sprintf(trace_buf, "load_keys_from_file: Password encrypt config: %s\n", password_encrypt_config);
		BFTRACE_BFCRYPT(trace_buf);
	}
	
	/* open the file, if not done above */
	f = fopen(password_encrypt_config, "r");

	/* return an error */
	if (f == NULL) {
		/* TODO: Need to link send_msg into the DLL w/o bringing in the kitchen sink.  
		The message is completed though.
		send_msg("PWCrypt", "PWCryptConfigFileNotFound", "s", password_encrypt_config);
		*/
		if (DEBUG) {
			sprintf(trace_buf, "load_keys_from_file: Could not open the file. Error: %d\n", errno);
			BFTRACE_BFCRYPT(trace_buf);
		}
		return errno;
	}
	buf = malloc(BFBUFSIZ);
	
	while (fgets(buf, BFBUFSIZ, f) != NULL) {
		
		/* If we get a starting point, create a new hashtable to add the keys to. */
		if (strstr(buf, "#****") != NULL) {
			ht = hash_new(ctx, 5);
			ht_list[num_keys] = ht;
			num_keys++;
			continue;
		}
		
		/* Parsing the first '=' only, other '=' might be part of base64.
		   Of course, there's probably an easier way. */
		total_size=strlen(buf);
		separator_ptr = strstr(buf, "=");
		if (separator_ptr != NULL) {
			key_size=(separator_ptr-buf)+1;
			key = malloc(sizeof(char)*(key_size));
			memcpy(key, buf, key_size);
			key[key_size-1]='\0';
			if (key[0] == '\0' || key[0] == '#' || key[0] == '\\') continue;
			
			value_size=total_size-key_size;  /* don't count the separating = and \n */
			value = malloc(sizeof(char)*(value_size+1));  /* add one for the trailing 0 */
			memcpy(value, buf+key_size, value_size);
			value[value_size-1]='\0';
			if (DEBUG) {
				sprintf(trace_buf, "load_keys_from_file: Value: %s, Value length: %d, strlen: %d\n", value, value_size, strlen(value));
				BFTRACE_BFCRYPT(trace_buf);
			}
			
			if (hash_contains_key(ht, key, key_size)) continue;

			if (value != NULL) {
				for (i=0; i<(signed)strlen(value); i++) {
					if (value[i] == '\n' || value[i] == '\r')
						value[i] = '\0';
				}
			}

			hash_put(ht, key, key_size, value);
		}
	}

	if (DEBUG) {
		for (j=0; j<num_keys; j++) {
			ht = ht_list[j];
			BFTRACE_BFCRYPT("load_keys_from_file: key list: \n");
			for (i=0; i<ht->size; i++) {
				sprintf(trace_buf, "  %3d:", i);
				BFTRACE_BFCRYPT(trace_buf);
				list = ht->data[i];
				while (list != NULL) {
					if (strstr((char *)list->key, "password") == NULL) {
						sprintf(trace_buf, strings ? "  %s-> \"%s\"" : " %s-> %p",
								(char *)list->key,
								(char *)list->value);
						BFTRACE_BFCRYPT(trace_buf);
					}
					list = list->next;
				}
				BFTRACE_BFCRYPT("\n");
			}
		}
	}

	if (DEBUG) {
		sprintf(trace_buf, "load_keys_from_file: Parsed %d key configurations.", num_keys);
		BFTRACE_BFCRYPT(trace_buf);
	}
	free(buf);
	fclose(f);
	return 0;
}

#ifdef WIN32
__declspec(dllexport) passwd_ops* bfcrypt_init(char * params);
#endif
passwd_ops* bfcrypt_init(char * params) {
	int rc=0;
	char *p;
	
    p = getenv ( "BFDEBUG_SECURITY" );
    if (p != NULL) {
        DEBUG=1;
    }

	rc = load_keys_from_file(params);
	if (rc != 0) {
		if (DEBUG) {
			sprintf(trace_buf, "Error loading key file.  Error = %d", rc);
			BFTRACE_BFCRYPT(trace_buf);
		}
		return NULL;
	}
  
    return &my_ops;
}
