/*
Copyright (C)2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#ifdef WIN32
	#include <excpt.h>
	#include <strsafe.h>
#else
	#include <dlfcn.h>
#endif

static const char* CRYPTO_ALGORITHM_STARTED = "{";
static const char* CRYPTO_ALGORITHM_STOPPED = "}";
static const char* CRYPTO_ALGORITHM_SEPARATOR = ":";
static const char MODULE_NAME[]="bfcrypt.dll";
static const char* PASSWORD_ENCRYPT_MODULE_PROPERTY="password_encrypt_module";
static passwd_ops* myops=NULL;
static int initialized = 0;
static int DEBUG=1;
static char trace_buf[512];

#ifdef WIN32
typedef passwd_ops* (*InitializationProcedure)(char *);
#else 
typedef passwd_ops* (*InitializationProcedure)(char *);
#endif


#ifndef WIN32
static void *dllhandle = NULL;

static int unix_loader_init(void) {
	InitializationProcedure functionhandle;
	char *params = NULL, *dll_path = NULL, *bfpwcrypt_path = NULL;
	char *function_name = NULL;
	char *temp = NULL;
	char *word, *lastword, *module_name, *error;
	const char *sep;
	int ret = 0;

	/* Already initialized or no module is specified in the conf file */
	if (initialized) { return 0; }
	params = bfconf_get(PASSWORD_ENCRYPT_MODULE_PROPERTY, NULL);
	if (NULL == params) { return 0; }

	/* Remove the path to get just "split" */
	temp = strdup(params);
	sep = ";";
	dll_path = strtok(temp, sep);
	bfpwcrypt_path = strtok(NULL, sep);
	if (NULL != activity_file && bfpwcrypt_path != NULL) {
		activity_write(BFHERE "bfpwcrypt_path: [%s]", bfpwcrypt_path);
	}
	
	/* Remove the path to get just "module.dll" */
	temp = strdup(dll_path);
	sep = "\\/";
	lastword = NULL;
	for (word = strtok(temp, sep); word; word = strtok(NULL, sep)) {
		lastword = word;
	}

	/* If dll_path is NULL or matches the regular expression /^[\/\\]*$/,
	 * then lastword will be NULL, here.
	 */
	if (NULL == lastword) {
		send_msg("AUTH", "AuthModuleConfMalformed", NULL);
		ret = -1;
		goto done;
	}
			
	/* Now remove ".dll" from the end. */
	module_name = strtok(lastword, ".");

	/* If dll_path is matches the regular expression /^[\/\\]*\.[\.\/\\]*$/,
	 * then module_name will be NULL, here.
	 */
	if (NULL == module_name) {
		send_msg("AUTH", "AuthModuleConfMalformed", NULL);
		ret = -2;
		goto done;
	}
	if (NULL != activity_file) {
		activity_write(BFHERE "module [%s]", module_name);
		if (!strcmp(module_name, "bfcrypt")) {
			activity_write(BFHERE "Debugging info for password decrypt operations can be found in ./bfcrypt.txt.");
		}
	}

	/* This allows us to load "module_init" per the struct design. */
	function_name = malloc(strlen(module_name) + 6);
	strcpy(function_name, module_name);
	strcat(function_name, "_init");
	if (NULL != activity_file) {
		activity_write(BFHERE "loading [%s]", dll_path);
	}

	/* Try to open the .so library configured. */
	dllhandle = dlopen(dll_path, RTLD_NOW);
	if (!dllhandle) {
		dllhandle = dlopen(dll_path, RTLD_GLOBAL| RTLD_LAZY);
		if (!dllhandle) {
			error = dlerror();
			ret = -3;
			if (NULL != activity_file && strcmp(module_name, "bfcrypt") == 0) {
				activity_write(BFHERE "Ensure bfpwcrypt.conf has been added to the agent directory.  Search documentation for more information about bfpwcrypt.conf.");
			}
			send_msg("AUTH", "AuthModuleLoadFail", "iss",
					ret,
					module_name,
					(NULL != error) ? error : dll_path );
			goto done;
		}
	}
		
	/* Load the _init method. */
	functionhandle = (InitializationProcedure)dlsym(
			dllhandle, function_name);   
	if (!functionhandle) {
		error = dlerror();
		ret = -4;
		send_msg("AUTH", "AuthModuleNoInit", "iss",
				ret,
				module_name,
				(NULL != error) ? error : function_name );
		dlclose(dllhandle);
		dllhandle = NULL;
		goto done;
	}

	/* Call the function */
	if (NULL != activity_file) {
		activity_write(BFHERE "calling [%s: %s()]",
				dll_path, function_name);
	}
	myops = (*functionhandle)(bfpwcrypt_path);
	if (NULL == myops) {
		ret = -5;
		send_msg("AUTH", "AuthModuleInitFail", "iss",
				ret,
				module_name,
				function_name);
		dlclose(dllhandle);
		dllhandle = NULL;
		goto done;
	}

	initialized = 1;

done:
	if (NULL != temp) { free(temp); }
	if (NULL != function_name) { free(function_name); }
	if (NULL != dll_path) { free(dll_path); }
	return ret;
}
#endif


#ifdef WIN32
int filter(unsigned int code, struct _EXCEPTION_POINTERS *ep) {

	LPTSTR lpszFunction = TEXT("win_loader_init");
	LPVOID lpMsgBuf;
    LPVOID lpDisplayBuf;
    DWORD dw = GetLastError(); 

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    /* Display the error message and exit the process */

    lpDisplayBuf = (LPVOID)LocalAlloc(LMEM_ZEROINIT, (lstrlen((LPCTSTR)lpMsgBuf)+lstrlen((LPCTSTR)lpszFunction)+40)*sizeof(TCHAR)); 
	StringCchPrintf((LPTSTR)lpDisplayBuf, 
    LocalSize(lpDisplayBuf) / sizeof(TCHAR),
    TEXT("%s failed with error %d: %s"), 
    lpszFunction, dw, lpMsgBuf); 

	if (DEBUG) {
		sprintf(trace_buf, "Message = %s", lpDisplayBuf );
		BFTRACE(trace_buf);
		sprintf(trace_buf, "Code = %u", code );
		BFTRACE(trace_buf);
		sprintf(trace_buf, "Exception address = %u, code = %u, flags = %u, info = %u", ep->ExceptionRecord->ExceptionAddress, ep->ExceptionRecord->ExceptionCode, ep->ExceptionRecord->ExceptionFlags, ep->ExceptionRecord->ExceptionInformation );
		BFTRACE(trace_buf);
		sprintf(trace_buf, "Context flags = %u", ep->ContextRecord->ContextFlags );
		BFTRACE(trace_buf);
	}
	return 0;
} 
#endif


#ifdef WIN32
/**
 * This will load the DLL from the configuration.  If it's already been initialized, it should just return.
 **/
static int win_loader_init(void) {
	HMODULE dllhandle;
	InitializationProcedure  init_func;
	char function_name[256];
	char *params = NULL, *bfpwcrypt_path = NULL;
	char *module_name=NULL, *module_from_conf=NULL, *sep=NULL, *temp=NULL, *word=NULL, *lastword=NULL;
	
	__try { 
		if (DEBUG) {
			BFTRACE("Loading password encryption module.");
		}
		if (!initialized) {
			if (NULL != (params = bfconf_get(PASSWORD_ENCRYPT_MODULE_PROPERTY, NULL))) {

				/* Remove the path to get just "split" */
				temp = strdup(params);
				sep = ";";
				module_from_conf = strtok(temp, sep);
				bfpwcrypt_path = strtok(NULL, sep);
				if (DEBUG && bfpwcrypt_path != NULL) {
					sprintf(trace_buf, "bfpwcrypt_path: %s", bfpwcrypt_path);
					BFTRACE(trace_buf);
				}
	
				/* remove the path to get just "module.dll" */
				sep = "\\/:;=-";
				temp = strdup(module_from_conf);
				for (word = strtok(temp, sep); word; word = strtok(NULL, sep))
					lastword = word;
				
				/* now remove ".dll" from the end. */
				module_name = strtok(strdup(lastword), ".");
				if (DEBUG && module_name != NULL) {
					sprintf(trace_buf, "module: %s", module_name);
					BFTRACE(trace_buf);
				}

				/* this allows us to load "module_init" per the struct design. */
				strcpy(function_name, module_name);
				strcat(function_name, "_init");
				
				if (DEBUG) {
					sprintf(trace_buf, "Loading module: %s", module_from_conf);
					BFTRACE(trace_buf);
					if (strcmp(module_from_conf, "bfcrypt.dll") == 0) {
						sprintf(trace_buf, "Debugging info for bfcrypt.dll password decrypt operations can be found in ./bfcrypt.txt.");
						BFTRACE(trace_buf);
					}
				}

				/* load the library and initialize the function. */
				dllhandle = GetModuleHandle(module_from_conf);
				if (dllhandle == NULL) {
					dllhandle = LoadLibrary(module_from_conf);
				}

				if (dllhandle != NULL) {
					init_func = (InitializationProcedure)GetProcAddress(dllhandle, function_name);

					if (init_func == NULL) {
						send_msg("PWCrypt", "PWCryptFailedInitializingFunction", "si", function_name, GetLastError());
						if (DEBUG) {
							sprintf(trace_buf, "Failed initializing the function %s.  Error = %d", function_name, GetLastError());
							BFTRACE(trace_buf);
						}
						FreeLibrary(dllhandle);
						return GetLastError();
					}
				}
				else {
					send_msg("PWCrypt", "PWCryptFailedInitializingDLL", "si", module_from_conf, GetLastError());
					if (DEBUG) {
						sprintf(trace_buf, "Failed loading DLL, error code = %d", GetLastError());
						BFTRACE(trace_buf);
					}
					return GetLastError();
				}

				if (DEBUG) {
					sprintf(trace_buf, "Loading procedure %s.", function_name);
					BFTRACE(trace_buf);
				}

				/* Call the function */
				if ((myops = (passwd_ops*)init_func(bfpwcrypt_path)) == NULL) {
					if (DEBUG) {
						sprintf(trace_buf, "Error loading function: %d.", GetLastError());
						BFTRACE(trace_buf);
						if (strcmp(module_from_conf, "bfcrypt.dll") == 0) {
							sprintf(trace_buf, "Ensure bfpwcrypt.conf has been added to the agent directory.  Search documentation for more information about bfpwcrypt.conf.");
							BFTRACE(trace_buf);
						}
					}
					FreeLibrary(dllhandle);
					return GetLastError();
				}

				initialized = 1;
			} else {
				if (DEBUG) {
					sprintf(trace_buf, "Password encryption property %s is not configured.", PASSWORD_ENCRYPT_MODULE_PROPERTY);
					BFTRACE(trace_buf);
				}
			}
			if (module_from_conf != NULL) free (module_from_conf);
		}
	} __except(filter(GetExceptionCode(), GetExceptionInformation())){ 
		if (DEBUG) {
			BFTRACE("in except."); 
		}
	} 
	return 0;
}
#endif


char* createHeader(unsigned char *cipherdata, int cipherdata_length, char *module, char *info) {
	char *buffer, *cipherstring;
	int index=0, length=0;
	
	if (DEBUG) {
		sprintf(trace_buf, "Cipher length: %d", cipherdata_length);
		BFTRACE(trace_buf);
		sprintf(trace_buf, "Cipher data: %s", cipherdata);
		BFTRACE(trace_buf);
	}
	cipherstring = base64Encode(cipherdata, cipherdata_length);
	
	if (DEBUG) {
		sprintf(trace_buf, "Cipher string after encoding: %s", cipherstring);
		BFTRACE(trace_buf);
	}

	length = strlen(cipherstring)+strlen(info)+strlen(module)+strlen(CRYPTO_ALGORITHM_STARTED)+strlen(CRYPTO_ALGORITHM_SEPARATOR)+strlen(CRYPTO_ALGORITHM_STOPPED)+1;
	buffer = malloc(length);
	memset(buffer, 0, length);
	memcpy(buffer, CRYPTO_ALGORITHM_STARTED, strlen(CRYPTO_ALGORITHM_STARTED));
	index+=strlen(CRYPTO_ALGORITHM_STARTED);
	memcpy(buffer+index, module, strlen(module));
	index+=strlen(module);
	memcpy(buffer+index, CRYPTO_ALGORITHM_SEPARATOR, strlen(CRYPTO_ALGORITHM_SEPARATOR));
	index+=strlen(CRYPTO_ALGORITHM_SEPARATOR);
	memcpy(buffer+index, info, strlen(info));
	index+=strlen(info);
	memcpy(buffer+index, CRYPTO_ALGORITHM_STOPPED, strlen(CRYPTO_ALGORITHM_STOPPED));
	index+=strlen(CRYPTO_ALGORITHM_STOPPED);
	memcpy(buffer+index, cipherstring, strlen(cipherstring));
	/* now that it's into another buffer, we can free the memory allocated for ciphertext. */
	
	if (DEBUG) {
		sprintf(trace_buf, "Header: %s", buffer);
		BFTRACE(trace_buf);
	}

	return buffer;
}

/**
 * This method will use the plaintext as input and either encode or encrypt the data specified.   It will
 * encrypt the plaintext if the following uncommented line is found in bfagent.conf:   
 * #password_encrypt C:\\Agent\\bfcrypt.dll  
 * It's recommended to fully qualify the path to ensure loading the correct module for security reasons.
 * The ciphertext result returned should be freed after use.
 **/
int encrypt_pwd (const char* plaintext, char** ciphertext) {
	int rc=0, cipherdata_length=0;
	char *info=NULL, *module=NULL; 

	if (!initialized) { 
#ifdef WIN32
		rc = win_loader_init(); 
#else
		rc = unix_loader_init();
#endif
	}

	/* if the password encrypt module is not initialized, use bf_encrypt to encode passwords. */
	if (myops == NULL) {
		*ciphertext = bf_encrypt(plaintext);
		return rc;
	}
	
	/* call out to the encryption DLL, but need to break the header apart first. */
	rc = myops->encrypt_password(plaintext, (unsigned char**)ciphertext, &cipherdata_length, &info);
	
	if (ciphertext != NULL) {
		/* get the module name */
		module = myops->getName();
		if (module == NULL || strlen(module) == 0) {
			/* print_error(); */
			module="bfcrypt"; /* as long as we're consistent, this should work. */
		}
	
		/* add the header to the ciphertext */
		*ciphertext = createHeader((unsigned char*)*ciphertext, cipherdata_length, module, info);
	}

	return rc;
}

/**
 * Check that the implementation name is in the header of the ciphertext.
 **/
char* getNameFromHeader(const char* ciphertext) {
		/** Returning only "name" from {name:info}xxxx.  **/
		char *word, *module_name, *sep;

		if (ciphertext != NULL) {

			if (NULL == (word = strchr(ciphertext, '{'))) {
				return NULL;
			}

			sep = "{}";
			word = strtok(strdup(ciphertext), sep);

			if (word != NULL) {
				sep = ":";
				module_name = strtok(word, sep);

				if (module_name != NULL) {
					if (DEBUG) {
						sprintf(trace_buf, "module_name: %s", module_name);
						BFTRACE(trace_buf);
					}
					return module_name;
				}
			}
		}		
		return NULL;
}

/**
 * Check that the implementation info is in the header of the ciphertext.
 **/
char* getInfoFromHeader(const char* ciphertext) {
		/** Returning only "info" from {name:info}xxxx.  "info" is impl-specific but might be "key|cipher_length".  **/
		char *word, *module_name, *sep, *info;

		if (ciphertext != NULL) {
			sep = "{}";
			/* get "{" */
			word = strtok(strdup(ciphertext), sep);

			if (word != NULL) {
				sep = ":";
				module_name = strtok(word, sep);
				info = strtok(NULL, sep);

				if (info != NULL) {
					if (DEBUG) {
						sprintf(trace_buf, "custom info: %s", info);
						BFTRACE(trace_buf);
					}
					return info;
				}
			}
		}		
		return NULL;
}

/**
 * Get the base64Encoded cipherdata
 **/
unsigned char* getCipherDataFromHeader(const char* ciphertext, int* ciphertext_length) {
		/** Returning only data after }.  **/
		unsigned char *word;
		char *sep;
		
		if (ciphertext != NULL) {
			sep = "}\0";
			/* get "{" */
			word = (unsigned char*)strtok(strdup(ciphertext), sep);
			
			if (word != NULL) {
				/* get "data" after "}" */
				word = (unsigned char*)strtok(NULL, sep);
				if (DEBUG) {
					sprintf(trace_buf, "encoded/encrypted data=%s", word);
					BFTRACE(trace_buf);
				}
			}

			if (word != NULL) {
				word = base64Decode((char*)word, ciphertext_length);
				return word;
			}
		}		
		return NULL;
}


/**
 * Make sure the implementation name exists and matches the name returned by the crypto DLL.   If it does not match,
 * it's not going to be able to decrypt it.
 **/
int isValidName(char * module_name) {

	/* this is in case someone doesn't return data on getName(). */
	if (module_name == NULL || strlen(module_name) == 0) {
		return 0;
	}
	
	if (myops != NULL && (!strcmp(myops->getName(),module_name) || !strcmp("UNKNOWN", module_name))) {
		return 1;
	}

	return 0;
}

/**
 * Get the name from the header and see if it matches our loaded DLL.
 **/

int isPasswordEncrypted(const char* ciphertext) {
	char *name;
	name = getNameFromHeader(ciphertext);
	return isValidName(name);
}

/**
 * This method will use the ciphertext to determine if it's plaintext, encoded or encrypted.   If encrypted, it
 * needs to have the following uncommented line in bfagent.conf:   
 * #password_encrypt C:\\Agent\\bfcrypt.dll  
 * It's recommended to fully qualify the path to ensure loading the correct module for security reasons.
 * If encoded, then we will call bf_decrypt.   If plaintext, it will return the input after copying it to plaintext.
 * The plaintext result returned should be freed after use.
 **/
int decrypt_pwd (char** plaintext, const char* ciphertext) {
	unsigned char *cipherdata;
	char *module_name, *info, *temp;
	int rc=0, ciphertext_length=0;
	if (!initialized) { 
#ifdef WIN32
		rc = win_loader_init(); 
#else
		rc = unix_loader_init();
#endif
	}

	ciphertext_length = strlen(ciphertext);

	if (!isPasswordEncrypted(ciphertext)) {
		if ((module_name = getNameFromHeader(ciphertext)) != NULL) {
			/*
			 * this is an error condition... an encrypted password exists but
			 * it's not able to find the correct module to decrypt it.  Need
			 * to indicate we need a module name of "module_name".
			 */
			/* print_error(MODULE_NOT_FOUND); */
			plaintext = NULL;
			return -1;
		}

		temp = bf_decrypt(ciphertext);

		if (temp == NULL) {
			*plaintext = (char *)malloc(ciphertext_length+1);
			memset(*plaintext, 0, ciphertext_length+1);
			strcpy(*plaintext, ciphertext);
			BFTRACE("Password already in plaintext form.");
			return 0;
		} else {
			*plaintext = (char *)malloc(strlen(temp)+1);
			memset(*plaintext, 0, strlen(temp)+1);
			strcpy(*plaintext, temp);
			BFTRACE("Password decoded.");
			return 0;
		}
	}
	
	/* strip the header */
	info = getInfoFromHeader(ciphertext);
	cipherdata = getCipherDataFromHeader(ciphertext, &ciphertext_length);
	rc = myops->decrypt_password(plaintext, cipherdata, ciphertext_length, info);
	BFTRACE("Password decrypted.");
	return rc;
}

