/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

#define	PBUFSIZE	512
#define	KEYLEN	8
#define	MINLEN	16

/* returns a pointer to an allocated string with the decoded string */
char *bf_decrypt(const char *ct) {
	int i, len, ctlen;
	unsigned int value, offset;
	const char *p;
	char ptb[PBUFSIZE], ctb[PBUFSIZE];

	if (ct == NULL) return NULL;
	ctlen = strlen(ct) >> 1;
	if (ctlen < 25 || ctlen >= PBUFSIZE - 1) return NULL;
	for (i=0, p=ct; i<ctlen; i++, p+=2) {
		sscanf(p, "%02x", &value);
		ctb[i] = (char)value;
	}

	len = 255 - (ctb[8] & 0xFF);
	if (len != 255) {
		if (ctlen < 9 + len) return NULL;
		for (i=0; i<len; i++) {
			value = ctb[i+9] & 0xFF;
			if (value < 0x20 || value >= 0x7F) return NULL;
			offset = value + 0x011D - (ctb[i & 7] & 0xFF);
			ptb[i] = (char)(offset % 0x5F + 0x20);
		}
	} else {
		len = 255 - (ctb[9] & 0xFF);
		if (ctlen < 10 + len) return NULL;
		for (i=0; i<len; i++) {
			ptb[i] = (char)((ctb[i + 10] ^ ctb[i & 7]) & 0xFF);
			if (ptb[i] == '\0') return NULL;
		}
	}

	ptb[len] = '\0';
#ifdef ZOS
	convert_atoe_l(ptb, len);
#endif
	return strdup(ptb);
}

static const char *HEX = "0123456789abcdef";
char *bf_encrypt(const char *pt) {
	unsigned char k[8];
	unsigned char *buf, *p;
	int i, x, stime;
	long ltime;
	size_t len = strlen(pt);
#ifdef ZOS
	char *zp;
#endif

	if (len > 116) { return NULL; }
	p = buf = malloc((len << 1) + 64);

#ifdef ZOS
	zp = malloc(len + 1);
	memcpy(zp, pt, len + 1);
	pt = zp;
#endif

	/* seed rand() */
	ltime = time(NULL);
	stime = (unsigned) ltime/2;
	srand(stime);

	for (i=0; i<8; ++i) {
		k[i] = (unsigned char)rand();
		*p++ = HEX[(k[i] >> 4) & 0x0F];
		*p++ = HEX[k[i] & 0x0F];
	}
	*p++ = '0';
	*p++ = '0';
	x = 255 - (int)len;
	*p++ = HEX[(x >> 4) & 0x0F];
	*p++ = HEX[x & 0x0F];

	for (i=0; i<(int)len; i++) {
		int x = *pt++ ^ k[i & 7];
		*p++ = HEX[(x >> 4) & 0x0F];
		*p++ = HEX[x & 0x0F];
	}

	for (i=len; i<16; i++) {
		x = (int)rand();
		*p++ = HEX[(x >> 4) & 0x0F];
		*p++ = HEX[x & 0x0F];
	}

	*p = '\0';

#ifdef ZOS
	free(zp);
#endif
	return (char *)buf;
}

unsigned char* base64Decode (const char* data, int* data_length) {
#ifdef HAVE_SSL
    BIO *mem, *b64;
    int length=512;
    unsigned char *buffer=NULL;
	buffer = (unsigned char *)malloc(length);
    memset(buffer, 0, length);
    b64 = BIO_new(BIO_f_base64());
	BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
    mem = BIO_new_mem_buf((char*)data, strlen(data));
	mem = BIO_push(b64, mem);
	memset(buffer, 0, length);
	*data_length = BIO_read(mem, buffer, length);
	BIO_flush(mem);
	BIO_free_all(mem);
	return buffer;
#else 
	return NULL;
#endif
}

char* base64Encode (const unsigned char* data, const int data_length) {
#ifdef HAVE_SSL
    BIO *mem, *b64;
    char* outbuff;
	int output_size;
	
    outbuff = (char *)malloc(data_length);
    memset(outbuff, 0, data_length);
	b64 = BIO_new(BIO_f_base64());
    mem = BIO_new(BIO_s_mem());
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
	mem = BIO_push(b64, mem);
	BIO_write(mem, data, data_length);
    BIO_flush(mem);
	output_size = BIO_get_mem_data(mem, &outbuff);
	BIO_set_close(mem, BIO_NOCLOSE); 
	outbuff[output_size] = '\0';
	BIO_free_all(mem);
	return outbuff;
#else
	return NULL;
#endif
}

