/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

int RemoveFunction(char *, struct stat *, int);

int FuncRmdir(void)
{
	char *root, *dir, *odir, *userdata, *ccroot;
	char *buf = NULL;
	int ret = 0;
	int mapfail;
	size_t userdata_size;

	/*  locate the root element */
	if (NULL == (root = request_get(TAG_ROOT))) {
		send_missing(CMD_RMDIR, TAG_ROOT);
		return 1;
	}

	/*  locate the dir element */
	if (NULL == (odir = request_get(TAG_DIR))) {
		send_missing(CMD_RMDIR, TAG_DIR);
		return 2;
	}

	/* don't allow ..'s in dir */
	if (strstr(odir, "..")) {
		send_msg("RMDIR", "RmdirDotDot", NULL);
		return 3;
	}

	/*  locate the userdata element */
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);

	/*  locate the silent element */
	verbose = !request_has(TAG_SILENT);

	StoreUserData(userdata, userdata_size);

	mapfail = BFCONF_EXISTS(map_drive_is_failure);
	if ((platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail))
	{
		send_msg("MAP", "MapFail", NULL);
		FreeUserData();
		platform_unmap_drives(0);
		return 4;
	}

	ClobberPasswords();

	CleanPath(root);
	CleanPath(odir);

	root = env_expand_var(root, 0);
	dir = env_expand_var(odir, 0);
	CleanPath(dir);
	if (strcmp(dir, odir)) {
		if (verbose) {
			char *before = xlate_build_path(root, "", odir);
			char *after = xlate_build_path(root, "", dir);
			send_msg("RMDIR", "Xlate", "ss", before, after);
			free(after);
			free(before);
		}
	}

	if (verbose) { DumpEnv(); }

	ccroot = NULL;
	if (clearcase_view) {
		ccroot = calloc(1, BFBUFSIZ);
		if (BFCONF_EXISTS(cc_suppress_server_root)) {
			format_message(ccroot, "%s/%s", this_view(), clearcase_view);
		} else {
			format_message(ccroot, "%s/%s/%s", this_view(), clearcase_view, root);
		}
		root = ccroot;
	}

	/* build the full base path that we'll be removing */
	buf = malloc(strlen(root) + strlen(dir) + 8);
	format_message(buf, "%s/%s", root, dir);
	CleanPath(buf);

	/*  test to see if it already doesn't exist */
	if (0 == platform_access(buf, 0)) {
		/* buf now has the path to delete files from */
		/* the actual work is off in platform.c 'cause some systems don't have */
		/* nftw (?) and rm -rf doesn't work in windows */
		BFRemoveDirectory(buf);

		/*  test to see if it's still there */
		if (0 == platform_access(buf, 0)) { ret = 5; }
	}


	platform_unmap_drives(0);
	if (0 == ret) {
		BFTRACE("RMDIR ok");
		send_msg("RMDIR", "RmdirOk", "s", buf);
	} else {
		BFTRACE("RMDIR failed");
		send_msg("RMDIR", "RmdirFail", "s", buf);
	}
	FreeUserData();
	free(dir);
	free(buf);
	if (NULL != ccroot) { free(ccroot); }
	return ret;
}

