/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

static const char *sysconf_cpu_arch(platform_cpu_arch_t cpu_arch) {
	switch (cpu_arch) {
		case CPU_ARCH_UNKNOWN:              return NULL;
		case CPU_ARCH_ALPHA:                return "Alpha";
		case CPU_ARCH_HPPA:                 return "HP/PA";
		case CPU_ARCH_IA64:                 return "IA-64";
		case CPU_ARCH_MVS:                  return "MVS";
		case CPU_ARCH_PASE:                 return "PASE";
		case CPU_ARCH_PPC:                  return "PPC";
		case CPU_ARCH_S390:                 return "S390";
		case CPU_ARCH_SPARC:                return "SPARC";
		case CPU_ARCH_VAX:                  return "VAX";
		case CPU_ARCH_X86:                  return "X86";
	}
	return NULL;
}


static const char *sysconf_cpu_manufacturer(
		platform_cpu_manufacturer_t cpu_manu)
{
	switch (cpu_manu) {
		case CPU_MANU_UNKNOWN:              return NULL;
		case CPU_MANU_AMD:                  return "AMD";
		case CPU_MANU_APPLE:                return "Apple";
		case CPU_MANU_CYRIX:                return "Cyrix";
		case CPU_MANU_DEC:                  return "DEC";
		case CPU_MANU_HP:                   return "HP";
		case CPU_MANU_IBM:                  return "IBM";
		case CPU_MANU_INTEL:                return "Intel";
		case CPU_MANU_MOTOROLA:	            return "Motorola";
		case CPU_MANU_NEXGEN:	            return "NexGen";
		case CPU_MANU_NSC:	                return "National";
		case CPU_MANU_RISE:                 return "Rise";
		case CPU_MANU_SIS:                  return "SiS";
		case CPU_MANU_SUN:                  return "Sun";
		case CPU_MANU_TRANSMETA:            return "TransMeta";
		case CPU_MANU_UMC:                  return "UMC";
		case CPU_MANU_VIA:                  return "VIA";
	}
	return NULL;
}



static const char *sysconf_cpu_model(platform_cpu_model_t cpu_model) {
	switch (cpu_model) {
		case CPU_MODEL_UNKNOWN:             return NULL;
		case CPU_MODEL_IA64_ITANIUM:        return "Itanium";
		case CPU_MODEL_IA64_ITANIUM_2:      return "Itanium-2";
		case CPU_MODEL_PPC_6XX:             return "6xx";
		case CPU_MODEL_PPC_POWER:           return "POWER";
		case CPU_MODEL_PPC_RS64:            return "RS64";
		case CPU_MODEL_PPC_G3:              return "G3";
		case CPU_MODEL_PPC_G4:              return "G4";
		case CPU_MODEL_PPC_G5:              return "G5";
		case CPU_MODEL_PPC_CELL:            return "Cell";
		case CPU_MODEL_S390_S390:           return "S390";
		case CPU_MODEL_S390_S390X:          return "S390X";
		case CPU_MODEL_X86_386:             return "386";
		case CPU_MODEL_X86_486:             return "486";
		case CPU_MODEL_X86_586:             return "586";
		case CPU_MODEL_X86_686:             return "686";
		case CPU_MODEL_X86_64:              return "X86-64";
	}
	return NULL;
}


static void sysconf_cpu(char *buf) {
	const char *p;
	platform_sysconf_cpu_t *cpu = platform_sysconf_cpu_new();
	if (NULL == cpu)
		return;

	if (NULL != ( p = sysconf_cpu_arch(cpu->cpu_arch) )) {
		sprintf(buf, "CPU_ARCH %s", p);
		send_log(NULL, buf);
	}
	if (NULL != ( p = sysconf_cpu_manufacturer(cpu->cpu_manufacturer) )) {
		sprintf(buf, "CPU_MANUFACTURER %s", p);
		send_log(NULL, buf);
	}
	if (NULL != ( p = sysconf_cpu_model(cpu->cpu_model) )) {
		sprintf(buf, "CPU_MODEL %s", p);
		send_log(NULL, buf);
	}
	if (cpu->cpu_speed > 0) {
		sprintf(buf, "CPU_MHZ %ld", cpu->cpu_speed);
		send_log(NULL, buf);
	}
	if (cpu->cpu_load >= 0.0) {
		sprintf(buf, "CPU_LOAD %.3f", cpu->cpu_load);
		send_log(NULL, buf);
	}
	if (cpu->cpu_load1 >= 0.0) {
		sprintf(buf, "CPU_LOAD1 %.3f", cpu->cpu_load1);
		send_log(NULL, buf);
	}
	if (cpu->cpu_load5 >= 0.0) {
		sprintf(buf, "CPU_LOAD5 %.3f", cpu->cpu_load5);
		send_log(NULL, buf);
	}
	if (cpu->cpu_load15 >= 0.0) {
		sprintf(buf, "CPU_LOAD15 %.3f", cpu->cpu_load15);
		send_log(NULL, buf);
	}
	if (NULL != cpu->cpu_serial) {
		sprintf(buf, "CPU_SERIAL %s", cpu->cpu_serial);
		send_log(NULL, buf);
	}
	if (cpu->cpu_count > 0) {
		sprintf(buf, "NUM_CPU %d", cpu->cpu_count);
		send_log(NULL, buf);
	}
	platform_sysconf_cpu_destroy(cpu);
}

static void sysconf_disk(char *buf) {
	platform_sysconf_disk_t *disk;
	const char *dir;

	if (NULL == (dir = request_get(TAG_DIR)))
		return;

	disk = platform_sysconf_disk_new(dir);
	if (NULL == disk)
		return;

	if (disk->disk_free >= 0) {
		sprintf(buf, "DISK_FREE %.0f", disk->disk_free / DISK_REPORT_SIZE);
		send_log(NULL, buf);
	}
	if (disk->disk_total >= 0) {
		sprintf(buf, "DISK_TOTAL %.0f", disk->disk_total / DISK_REPORT_SIZE);
		send_log(NULL, buf);
	}
	platform_sysconf_disk_destroy(disk);
}

static void sysconf_mem(char *buf) {
	platform_sysconf_mem_t *mem;

	mem = platform_sysconf_mem_new();
	if (NULL == mem)
		return;
	if (mem->mem_page_size > 0) {
		sprintf(buf, "MEM_PAGESIZE %ld", mem->mem_page_size);
		send_log(NULL, buf);
	}
	if (mem->mem_total >= 0.0) {
		sprintf(buf, "MEM_TOTAL %.0f", mem->mem_total / MEM_REPORT_SIZE);
		send_log(NULL, buf);
	}
	if (mem->mem_free >= 0.0) {
		sprintf(buf, "MEM_FREE %.0f", mem->mem_free / MEM_REPORT_SIZE);
		send_log(NULL, buf);
	}
	if (mem->mem_load >= 0.0) {
		sprintf(buf, "MEM_LOAD %.0f", mem->mem_load * 100.0);
		send_log(NULL, buf);
	}

	platform_sysconf_mem_destroy(mem);
}

static void sysconf_net(char *buf) {
	platform_sysconf_net_t *net;
	net = platform_sysconf_net_new();
	if (NULL == net)
		return;
	if (NULL != net->net_fqdn) {
		sprintf(buf, "NET_FQDN %s", net->net_fqdn);
		send_log(NULL, buf);
	}
	if ('\0' != *net->net_hwaddr) {
		sprintf(buf, "NET_HWADDR %s", net->net_hwaddr);
		send_log(NULL, buf);
	}
	if (NULL != net->net_iface) {
		sprintf(buf, "NET_IFACE %s", net->net_iface);
		send_log(NULL, buf);
	}
	if (net->net_ipv > 0) {
		sprintf(buf, "NET_IPV %d", net->net_ipv);
		send_log(NULL, buf);
	}
	if ('\0' != *net->net_ipv4) {
		sprintf(buf, "NET_IPV4 %s", net->net_ipv4);
		send_log(NULL, buf);
	}
	if ('\0' != *net->net_ipv6) {
		sprintf(buf, "NET_IPV6 %s", net->net_ipv6);
		send_log(NULL, buf);
	}
	if (net->net_speed > 0) {
		sprintf(buf, "NET_SPEED %ld", net->net_speed);
		send_log(NULL, buf);
	}
	platform_sysconf_net_destroy(net);
}

static void sysconf_os(char *buf) {
	platform_sysconf_os_t *os = platform_sysconf_os_new();
	if (NULL != os->os_hostid) {
		sprintf(buf, "OS_HOSTID %s", os->os_hostid);
		send_log(NULL, buf);
	}
	if (NULL != os->os_sysname) {
		sprintf(buf, "OS_SYSNAME %s", os->os_sysname);
		send_log(NULL, buf);
	}
	if (NULL != os->os_version) {
		sprintf(buf, "OS_VERSION %s", os->os_version);
		send_log(NULL, buf);
	}
	if (NULL != os->os_release) {
		sprintf(buf, "OS_RELEASE %s", os->os_release);
		send_log(NULL, buf);
	}
	if (NULL != os->os_machine) {
		sprintf(buf, "OS_MACHINE %s", os->os_machine);
		send_log(NULL, buf);
	}
	if (os->os_win_service_pack > 0) {
		sprintf(buf, "WIN_SERVICEPACK %d", os->os_win_service_pack);
		send_log(NULL, buf);
	}
	platform_sysconf_os_destroy(os);
}



#define FLAG_CPU    ('C')
#define FLAG_DISK   ('D')
#define FLAG_MEM    ('M')
#define FLAG_NET    ('N')
#define FLAG_OS     ('O')

/*
 * Report information about the system
 *
 */
int FuncSysConf(void)
{
	char *buf = malloc(BFBUFSIZ);
	char *flags = request_get(TAG_FLAGS);

	ClobberPasswords();

	if (NULL == flags)
		flags = "CDMNO";

	if (NULL != strchr(flags, FLAG_CPU))
		sysconf_cpu(buf);
	if (NULL != strchr(flags, FLAG_DISK))
		sysconf_disk(buf);
	if (NULL != strchr(flags, FLAG_MEM))
		sysconf_mem(buf);
	if (NULL != strchr(flags, FLAG_NET))
		sysconf_net(buf);
	if (NULL != strchr(flags, FLAG_OS))
		sysconf_os(buf);

	send_msg("EXEC", "SysconfDone", NULL);
	free(buf);
	return 0;
}

