/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

int FuncUnlink(void)
{
	char *buf = NULL;
	char *root, *path, *opath, *userdata, *ccroot;
	int ret = 0;
	int mapfail;
	size_t userdata_size;

	/*  locate the root element */
	if (NULL == (root = request_get(TAG_ROOT))) {
		send_missing(CMD_UNLINK, TAG_ROOT);
		return 1;
	}

	/*  locate the path element */
	if (NULL == (opath = request_get(TAG_PATH))) {
		send_missing(CMD_UNLINK, TAG_PATH);
		return 2;
	}
	/* don't allow ..'s in path */
	if (strstr(opath, "..")) {
		send_msg("UNLINK", "UnlinkDotDot", NULL);
		return 3;
	}

	/*  locate the userdata element */
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);


	/*  locate the silent element */
	verbose = !request_has(TAG_SILENT);

	StoreUserData(userdata, userdata_size);

	mapfail = BFCONF_EXISTS(map_drive_is_failure);
	if ((platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail))
	{
		send_msg("MAP", "MapFail", NULL);
		FreeUserData();
		return 4;
	}

	ClobberPasswords();

	CleanPath(root);
	CleanPath(opath);

	root = env_expand_var(root, 0);
	path = env_expand_var(opath, 0);
	CleanPath(path);
	if (strcmp(path, opath)) {
		if (verbose) {
			char *before = xlate_build_path(root, "", opath);
			char *after = xlate_build_path(root, "", path);
			send_msg("UNLINK", "Xlate", "ss", before, after);
			free(after);
			free(before);
		}
	}

	ccroot = NULL;
	if (clearcase_view) {
		ccroot = calloc(1, BFBUFSIZ);
		if (BFCONF_EXISTS(cc_suppress_server_root)) {
			format_message(ccroot, "%s/%s", this_view(), clearcase_view);
		} else {
			format_message(ccroot, "%s/%s/%s", this_view(), clearcase_view, root);
		}
		root = ccroot;
	}

	/* build the full base path that we'll be removing */
	format_message(buf, "%s/%s", root, path);
	CleanPath(buf);

	if (verbose) { DumpEnv(); }

	/*  test to see if it doesn't exist already */
	if (0 == platform_access(buf, 0)) {
		/* check to see if it's a dir */
		if (isDir(buf)) {
			send_msg("UNLINK", "UnlinkDir", "s", buf);
			ret = 5;
		} else {
			/* all good. blow it away */
			platform_unlink(buf);

			/*  test to see if it's still there */
			ret = (0 == platform_access(buf, 0)) ? 6 : 0;
		}
	}

	platform_unmap_drives(0);
	if (ret == 0) {
		BFTRACE("UNLINK ok");
		send_msg("UNLINK", "UnlinkOk", "s", buf);
	} else {
		BFTRACE("UNLINK failed");
		send_msg("UNLINK", "UnlinkFail", "s", buf);
	}
	FreeUserData();
	free(path);
	if (ccroot) { free(ccroot); }
	return ret;
}

