/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

int FuncWrite(void)
{
	char	*buf = malloc(BFBUFSIZ), *path, *p, *userdata;
	char	*timeout;
	int 	bytes, total, len, fd = -1, mode, append, clobber;
	time_t	stamp;
	struct 	utimbuf tims;
	int 	mapfail, rc = 0, arg = -1;
	int     stream_fd = -1;
	off_t   offset = (off_t)0;
	size_t	userdata_size;

	/*  locate the path element */
	if ((path = request_get(TAG_PATH)) == NULL) {
		send_missing(CMD_WRITE, TAG_PATH);
		free(buf);
		return 1;
	}

	stamp = 0;
	if (NULL != (p = request_get(TAG_STAMP)))
		stamp = atol(p);

	/* See if a data stream was provided */
	stream_fd = request_get_stream(TAG_STREAM);
	if (-1 == (stream_fd = request_get_stream(TAG_STREAM))) {
		send_missing(CMD_WRITE, TAG_STREAM);
		free(buf);
		return 2;
	}

	/* We should not see this tag anymore, but support it anyway... */
	p = request_get(TAG_OFFSET);
	if (NULL != p) {
		while (isdigit((int)*p)) {
			offset = offset * 10 + (*p++ - '0');
		}
		if (offset < 0) { offset = 0; }
	}

	/*  locate the append element. if it wasn't supplied, set a default. */
	append = request_has(TAG_APPEND) ? O_APPEND : 0;

	/*  locate the clobber element. if it wasn't supplied, set a default. */
	clobber = request_has(TAG_CLOBBER);

	/*  locate the mode element. if it wasn't supplied, set a default. */
	mode = -1;
	if (NULL != (p = request_get(TAG_MODE))) {
		sscanf(p, "%o", (unsigned int *)&mode);
	}

	/*  locate the userdata element */
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);

	/*  locate the client timeout element */
	if (NULL == (timeout = request_get(TAG_TIMEOUT)))
		timeout = "0";

	StoreUserData(userdata, userdata_size);

	mapfail = BFCONF_EXISTS(map_drive_is_failure);
	if ((platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail))
	{
		send_msg("MAP", "MapFail", NULL);
		FreeUserData();
		free(buf);
		platform_unmap_drives(0);
		return 3;
	}

	ClobberPasswords();

	if (clearcase_view) {
		char *view = this_view();
		p = calloc(1, strlen(view) + strlen(clearcase_view) +
				strlen(path) + 3);
		format_message(p, "%s/%s/%s", view, clearcase_view, path);
		path = p;
	} else {
		path = strdup(path);
	}
	/* Note: Do not use "goto done;" to bail out until after this
	 * point, as it would try to free(path);
	 */

	/* process the clearcase view/vob start/mount functions */
	CCStartViews(atoi(timeout));
	CCMountVobs(atoi(timeout));

	/* assemble and clean the final pathname. clean the
	 * pieces so we can use them individually later. */
	CleanPath(path);
	path = env_expand_var((p = path), 0);
	free(p);

	/* clobber the target file?
	 * TODO: Error handling would be nice...
	 */
	if (clobber)
		platform_unlink(path);

	/* open the target file */
	total = 0;
	if (!clobber && 0 == platform_access(path, 0)) {
		if ((fd = platform_open2(path, O_WRONLY|O_BINARY|append)) >= 0) {
			if (offset > 0) {
				lseek(fd, offset, SEEK_SET);
			}
		}
	} else {
		fd = platform_open3(path, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY|append, 0600);
	}
	if (fd < 0) {
		send_msg("WRITE", "WriteNoOpen", "s", path);
		rc = 4;
		arg = -3;
		goto done;
	}

	lseek(stream_fd, 0, SEEK_SET);
	while ((len = read(stream_fd, buf, BFBUFSIZ)) > 0) {
		char *p = buf;
		while (len > 0) {
			bytes = write(fd, p, len);
			if (bytes <= 0) {
				BFTRACE("WRITE failed");
				send_msg("WRITE", "WriteFail", "si", path, errno);
				lseek(stream_fd, 0, SEEK_END);
				rc = 6;
				goto done;
			}
			total += bytes;
			p += bytes;
			len -= bytes;
		}
	}
	close(fd);
	fd = -1;

	/* Set the file's permissions if a specific value was requested. */
	if (mode != -1) {
		if (0 == platform_chmod(path, mode)) {
			send_msg("WRITE", "WriteSetMode", "si",
					path, mode);
		} else {
			send_msg("WRITE", "WriteSetModeFail", "sii",
					path, mode, errno);
		}
	}

	/* Set the file's timestamp if a specific value was requested. */
	if (stamp > 0) {
		tims.actime = tims.modtime = stamp;
		if (0 == platform_utime(path, &tims)) {
			send_msg("WRITE", "WriteSetTime", "ss",
					path, ctime(&stamp));
		} else {
			send_msg("WRITE", "WriteSetTimeFail", "ssi",
					path, ctime(&stamp), errno);
		}
	}

	BFTRACE("WRITE ok");
	send_msg("WRITE", "WriteOk", "is", total, path);

	/* Note: See warning above about the point at which 'goto done;' is
	 * safe.
	 */
done:
	if (fd >= 0) { close(fd); }
	platform_unmap_drives(0);
	FreeUserData();
	free(path);
	free(buf);
	return rc;
}

